/*
 * Dombey Copyright (c) 2010, 2011, James Bailie.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *     * The name of James Bailie may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Multi-process SCGI application server as a library.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/event.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <sys/un.h>

#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>

#include <pwd.h>
#include <grp.h>
#include <signal.h>

#include <syslog.h>
#include <stdarg.h>
#include <string.h>

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

#define VERSION "2.1"

#define NOT_BUSY 0
#define BUSY 1
#define EMPTY -1

#define MAX_ENV 50
#define MAX_PARAM 50
#define MAX_COOKIE 50
#define MAX_FILE 75

extern char *optarg;

int scgi_fd = -1, scgi_testing = 0, scgi_backlog = 1000, scgi_min_slaves = 3,
   scgi_max_body = 10000, scgi_logging = 0, scgi_max_slaves = 25,
   scgi_max_file = 2000000, scgi_slave_no = -1, scgi_header_len = 0,
   scgi_interval = 0, scgi_counter = 0;

void ( *scgi_periodic_func )() = NULL;

volatile int scgi_killed = 0;

pid_t *scgi_slaves = NULL;

char *scgi_interface = "", *scgi_port = "4000", *scgi_grp = "nobody",
   *scgi_user = "nobody", *scgi_busymap = NULL, *scgi_header = NULL,
   *scgi_body = NULL, *scgi_root_dir = NULL, *scgi_listen_unix = NULL,

   *scgi_env[ MAX_ENV + 1 ], *scgi_params[ MAX_PARAM + 1 ],
   *scgi_cookies[ MAX_COOKIE + 1 ], *scgi_files[ MAX_FILE + 1 ],

   *scgi_config_file = NULL,
   *scgi_query_string = NULL, *scgi_http_cookie = NULL,
   *scgi_app_name = "LibDombey", *scgi_pidfile = "LibDombey.pid";

struct passwd *scgi_passwd;
struct group *scgi_group;

char *scgi_get_env( char * );
char **scgi_get_envs();

char *scgi_get_param( char * );
char **scgi_get_params();

char *scgi_get_cookie( char * );
char **scgi_get_cookies();

void scgi_set_name( char * );
void scgi_init_func();
void scgi_child_init_func();

void scgi_request_handler();
char *scgi_form_encode( char * );

int scgi_stack_inc = 128;

struct scgi_string
{
   int free, used;
   char *top;
   char *str;
};

void *scgi_memory( int size )
{
   void *ptr;

   if ( size == 0 )
      return NULL;

   if (( ptr = malloc( size )) == NULL )
   {
      if ( scgi_logging )
         syslog( LOG_WARNING, "malloc(): %m" );
      else
         fprintf( stderr, "dombey: malloc(): %s\n", strerror( errno ));

      exit( 1 );
   }

   return ptr;
}

char *scgi_str_dup( char *str, int len )
{
   char *ptr;

   if ( len < 0 )
      len = strlen( str );

   ptr = ( char *)scgi_memory( len + 1 );
   bcopy( str, ptr, len );
   ptr[ len ] = '\0';

   return ptr;
}

struct scgi_string *scgi_make_string()
{
   struct scgi_string *s;

   s = ( struct scgi_string *)scgi_memory( sizeof( struct scgi_string ));
   s->str = ( char *)scgi_memory( scgi_stack_inc + 1 );
   *s->str = '\0';
   s->free = scgi_stack_inc;
   s->used = 0;
   s->top = s->str;

   return s;
}

void scgi_string_append( struct scgi_string *s, char c )
{
   if ( s->free == 0 )
   {
      s->str = ( char *)realloc( s->str, s->used + 1 + scgi_stack_inc );

      if ( s->str == NULL )
      {
         if ( scgi_logging )
            syslog( LOG_WARNING, "realloc(): %m" );
         else
            fprintf( stderr, "dombey: realloc(): %s\n", strerror( errno ));
      }

      /* Leave room for end-of-string sentinel */

      s->free = scgi_stack_inc;
      s->top = &s->str[ s->used ];
   }

   ++s->used;
   --s->free;
   *s->top++ = c;
   *s->top = '\0';
}

#define STRING_APPEND( _s_, _c_ ) { if (( _s_ )->free )\
   {\
      --( _s_ )->free;\
      ++( _s_ )->used;\
      *( _s_ )->top++ = _c_;\
      *( _s_ )->top ='\0';\
   }\
   else scgi_string_append( _s_, _c_ ); }

void scgi_string_truncate( struct scgi_string *s )
{
   if ( s->used )
   {
      s->free += s->used;
      s->used = 0;
      *s->str = '\0';
      s->top = s->str;
   }
}

char *scgi_form_decode( char *str )
{
   static struct scgi_string *s = NULL;
   char *ptr;

   if ( s == NULL )
      s = scgi_make_string();
   else
      scgi_string_truncate( s );

   for( ptr = str; *ptr; ++ptr )
      switch( *ptr )
      {
         case '+':
            STRING_APPEND( s, ' ' )
            break;

         case '%':
            if ( ! *( ptr + 1 ) || ! *( ptr + 2 ))
               STRING_APPEND( s, *ptr )
            else
            {
               char data[ 3 ];
               char code;

               data[ 0 ] = *( ptr + 1 );
               data[ 1 ] = *( ptr + 2 );
               data[ 2 ] = '\0';

               if (( code = ( char )strtol( data, NULL, 16 )))
               {
                  STRING_APPEND( s, code )
                  ptr += 2;
               }
               else
                  STRING_APPEND( s, *ptr )
            }
            break;

         default:
            STRING_APPEND( s, *ptr )
      }

   return scgi_str_dup( s->str, s->used );
}

char *scgi_form_encode( char *str )
{
   char *ptr, *ptr2;
   static struct scgi_string *s = NULL;
   int found;
   static char *reserved = "]$;:@&+!?=#{}/[^`~\"<>|%\\\t\r\n",
               *letters  = "0123456789ABCDEF";

   if ( s == NULL )
      s = scgi_make_string();
   else
      scgi_string_truncate( s );

   for( ptr = str; *ptr; ++ptr )
   {
      found = 0;

      if ( *ptr == ' ' )
      {
         STRING_APPEND( s, '+' )
         continue;
      }

      if ( *ptr < 32 || *ptr > 126 )
      {
         STRING_APPEND( s, '%' )
         STRING_APPEND( s, letters[ *ptr / 16 ] )
         STRING_APPEND( s, letters[ *ptr % 16 ] )
         continue;
      }

      for( ptr2 = reserved; *ptr2; ++ptr2 )
         if ( *ptr == *ptr2 )
         {
            STRING_APPEND( s, '%' )
            STRING_APPEND( s, letters[ *ptr / 16 ] )
            STRING_APPEND( s, letters[ *ptr % 16 ] )
            found = 1;
            break;
         }

      if ( ! found )
         STRING_APPEND( s, *ptr )
   }

   return scgi_str_dup( s->str, s->used );
}

/*
 * If we are a slave process, we set the killed flag, and when we are
 * finished servicing any current connection, we will exit.  It would be
 * nice to unlink the master process's pidfile here, but we no longer are
 * root and so cannot access /var/run/
 */

void scgi_signal_handler( int signo )
{
   int i;

   if ( scgi_slave_no >= 0 )
      scgi_killed = 1;
   else
   {
      for( i = 0; i < scgi_max_slaves; ++i )
         if ( scgi_slaves[ i ] )
            kill( scgi_slaves[ i ], SIGTERM );

      exit( 0 );
   }
}

/*
 * x - Do not become daemon.  Run single process iterative server.

 * r - root directory
 * l - socket for unix domain server
 * m - max slaves
 * n - min free slaves
 * p - port
 * f - config file path
 * i - interface
 * u - user
 * g - group
 * b - max size of request bodies
 * e - max size of uploaded files
 * f - config file
 */

void scgi_set_options( int argc, char **argv )
{
   int i, specified = 0;

   while(( i = getopt( argc, argv, "xe:l:f:r:m:b:n:p:i:u:g:")) != -1 )
      switch( i )
      {
         case '?':
            exit( 1 );

         case 'b':
            scgi_max_body = strtol( optarg, NULL, 10 );
            break;

         case 'e':
            scgi_max_file = strtol( optarg, NULL, 10 );
            break;

         case 'f':
            /* already set in scgi_init_config_file */
            break;

         case 'r':
            scgi_root_dir = optarg;
            break;

         case 'g':
            scgi_grp = optarg;
            break;

         case 'i':
            ++specified;
            scgi_interface = optarg;
            break;

         case 'l':
            scgi_listen_unix = optarg;
            break;

         case 'm':
            scgi_max_slaves = strtol( optarg, NULL, 10 );
            break;

         case 'n':
            scgi_min_slaves = strtol( optarg, NULL, 10 );
            break;

         case 'p':
            ++specified;
            scgi_port = optarg;
            break;

         case 'u':
            scgi_user = optarg;
            break;

         case 'x':
            scgi_testing = 1;
      }

   if ( specified && scgi_listen_unix != NULL )
   {
      fprintf( stderr,
               "%s: the -l option cannot be present when either or both of the -i and -p options are present",
               scgi_app_name );
      exit( 1 );
   }

   if ( scgi_root_dir == NULL )
   {
      fprintf( stderr, "%s: root directory undefined (missing -r option)", scgi_app_name );
      exit( 1 );
   }
   
   if ( chdir( scgi_root_dir ) < 0 )
   {
      fprintf( stderr, "%s: chdir( %s ): %s\n",
               scgi_app_name, scgi_root_dir, strerror( errno ));
      exit( 1 );
   }

   if ( scgi_min_slaves <= 0 )
   {
      fprintf( stderr, "%s: min slaves <= 0: %d\n", scgi_app_name,
               scgi_min_slaves );
      exit( 1 );
   }

   if ( scgi_max_slaves <= 0 )
   {
      fprintf( stderr, "%s: max slaves <= 0: %d\n", scgi_app_name,
               scgi_max_slaves );
      exit( 1 );
   }

   if ( scgi_min_slaves > scgi_max_slaves )
   {
      fprintf( stderr, "%s: min slaves > max_slaves: %d > %d\n",
               scgi_app_name, scgi_min_slaves, scgi_max_slaves );
      exit( 1 );
   }

   if ( scgi_max_body <= 0 )
   {
      fprintf( stderr, "%s: max_body <= 0: %d\n", scgi_app_name,
               scgi_max_body );
      exit( 1 );
   }

   if (( scgi_passwd = getpwnam( scgi_user )) == NULL )
   {
      fprintf( stderr, "%s: user \"%s\" does not exist\n", scgi_app_name,
               scgi_user );
      exit( 1 );
   }

   if (( scgi_group = getgrnam( scgi_grp )) == NULL )
   {
      fprintf( stderr, "%s: group \"%s\" does not exist\n", scgi_app_name,
               scgi_grp );
      exit( 1 );
   }
}

void scgi_become_daemon()
{
   int file;

   /*
    * Fork and let the parent die, continuing as child so we are not
    * a process group leader.  This is necessary for the call to setsid().
    */

   switch( fork() )
   {
      case -1:
         fprintf( stderr, "%s: fork(): %s\n", scgi_app_name,
                  strerror( errno ));
         exit( 1 );

      case 0:
         break;

      default:
         exit( 0 );
   }

   fclose( stdout );
   fclose( stderr );
   fclose( stdin );

   stdin = fopen( "/dev/null", "r" );
   stdout = fopen( "/dev/null", "w" );
   stderr = fopen( "/dev/null", "w" );

   if ( stdin == NULL || stdout == NULL || stderr == NULL )
   {
      syslog( LOG_ERR, "fopen(): %m" );
      exit( 1 );
   }

   /*
    * Detach us from our controlling terminal, so job control and other
    * signals may not be sent to us from that terminal.
    */

   if ( setsid() < 0 )
   {
      syslog( LOG_ERR, "setsid(): %m" );
      exit( 1 );
   }

   /*
    * Write our pid to disk.
    */

   if ( ! scgi_testing )
   {
      if (( file = open( scgi_pidfile, O_WRONLY | O_CREAT | O_TRUNC,
                         S_IWUSR | S_IRUSR | S_IRGRP )) < 0 )
         syslog( LOG_WARNING, "open(): %m" );
      else
      {
         char buffer[ 16 ];

         snprintf( buffer, sizeof( buffer ), "%d", getpid() );
         write( file, buffer, strlen( buffer ));
         close( file );
      }
   }

   /*
    * We don't fork children which create files, but if we did, this will
    * make those files not world-readable by default.
    */

   umask( 2 );
}

void scgi_change_identity()
{
   if ( setgid( scgi_group->gr_gid ) < 0 )
      syslog( LOG_ERR, "setgid(): %m" );

   if ( setuid( scgi_passwd->pw_uid ) < 0 )
      syslog( LOG_ERR, "setuid(): %m" );
}

void scgi_close_on_exec( int fd )
{
   if ( fcntl( fd, F_SETFD, FD_CLOEXEC ) < 0 )
      syslog( LOG_ERR, "fcntl( F_SETFD, FD_CLOEXEC ): %m" );
}

void scgi_start_listening_unix()
{
   struct sockaddr_un sa;

   if (( scgi_fd = socket( PF_LOCAL, SOCK_STREAM, 0 )) < 0 )
   {
      syslog( LOG_ERR, "socket(): %m" );
      exit( 1 );
   }

   unlink( scgi_listen_unix );
   bzero( &sa, sizeof( struct sockaddr_un ));
   strncpy( sa.sun_path, scgi_listen_unix, sizeof( sa.sun_path ) - 1 );  /* ensures NUL-terminated. */

   if ( bind( scgi_fd, ( struct sockaddr *)&sa, SUN_LEN( &sa )))
   {
      syslog( LOG_ERR, "bind( %s ): %m", scgi_listen_unix );
      close( scgi_fd );
      exit( 1 );
   }

   if ( chown( scgi_listen_unix, scgi_passwd->pw_uid, scgi_group->gr_gid ) < 0 )
   {
      syslog( LOG_ERR, "chown( %s ): %m", scgi_listen_unix );
      close( scgi_fd );
      exit( 1 );
   }

   if ( chmod( scgi_listen_unix, S_IRWXU | S_IRWXG ) < 0 )
   {
      syslog( LOG_ERR, "chmod( %s, S_IRWXU | S_IRWXG ): %m", scgi_listen_unix );
      close( scgi_fd );
      exit( 1 );
   }
   
   if ( listen( scgi_fd, scgi_backlog ) < 0 )
   {
      syslog( LOG_ERR, "listen(): %m" );
      close( scgi_fd );
      exit( 1 );
   }

   scgi_close_on_exec( scgi_fd );
}

void scgi_start_listening()
{
   struct addrinfo hints, *res;
   int result;

   if ( scgi_listen_unix != NULL )
   {
      scgi_start_listening_unix();
      return;
   }

   /*
    * This is the new, protocol-independent way of setting up a listening
    * socket.
    */

   bzero( &hints, sizeof( struct addrinfo ));
   hints.ai_flags = AI_PASSIVE;
   hints.ai_socktype = SOCK_STREAM;

   /*
    * If the user has not specified an interface, we listen on the
    * IPv6 wildcard address.
    */

   if (( result = getaddrinfo( ( *scgi_interface ? scgi_interface : NULL ),
                               scgi_port, &hints, &res )))
   {
      syslog( LOG_ERR, "getaddrinfo(): %s", gai_strerror( result ));
      exit( 1 );
   }

   if ( res == NULL )
   {
      syslog( LOG_ERR, "getaddrinfo(): no interface found" );
      exit( 1 );
   }

   scgi_fd = socket( res->ai_family, res->ai_socktype, res->ai_protocol );

   if ( scgi_fd == -1 )
   {
      syslog( LOG_ERR, "socket(): %m" );
      exit( 1 );
   }

   result = 1;

   /*
    * Allow duplicate bindings, so we can have more than one
    * server process, if necessary.
    */

   if ( setsockopt( scgi_fd, SOL_SOCKET, SO_REUSEADDR, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_REUSEADDR ): %m" );

   /*
    * Try and detect connections which go idle for long periods of time.
    */

   if ( setsockopt( scgi_fd, SOL_SOCKET, SO_KEEPALIVE, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( SO_KEEPALIVE ): %m" );

   /*
    * Make sure we can accept IPv4 traffic as well as IPv6, if we
    * are bound to the IPv6 wildcard address.  We are turning off
    * the option because result = 0.
    */

   result = 0;

   if ( ! *scgi_interface &&
        setsockopt( scgi_fd, IPPROTO_IPV6, IPV6_BINDV6ONLY, &result,
                    sizeof( result )) < 0 )
      syslog( LOG_WARNING, "setsockopt( IPV6_BINDV6ONLY ): %m" );

   if ( bind( scgi_fd, res->ai_addr, res->ai_addrlen ) < 0 )
   {
      syslog( LOG_ERR, "bind(): %m" );
      exit( 1 );
   }

   if ( res != NULL )
      freeaddrinfo( res );

   if ( listen( scgi_fd, scgi_backlog ) < 0 )
   {
      syslog( LOG_ERR, "listen(): %m" );
      exit( 1 );
   }

   scgi_close_on_exec( scgi_fd );
}

/*
 * If we are a slave, and we receive a SIGTERM, we will exit at our next
 * convenience, either the next time accept_connection() is called, after
 * we have closed the previous connection, but before we accept another, or
 * if we are already blocked in accept(), immediately, when accept() is
 * interrupted by the signal.
 */

void scgi_accept_connection()
{
   struct sockaddr_storage addr;
   socklen_t len;
   int conn;

   fclose( stdin );
   fclose( stdout );

AGAIN:
   if ( scgi_killed )
      exit( 0 );

   len = sizeof( struct sockaddr_storage );
   conn = accept( scgi_fd, ( struct sockaddr *)&addr, &len );

   if ( conn < 0 )
   {
      if ( errno == EINTR )
         goto AGAIN;

      syslog( LOG_ERR, "accept(): %m" );
      exit( 1 );
   }

   scgi_close_on_exec( conn );

   if (( stdin = fdopen( conn, "r" )) == NULL )
   {
      syslog( LOG_ERR, "fdopen(): %m" );
      exit( 1 );
   }

   if (( conn = dup( conn )) < 0 )
   {
      syslog( LOG_ERR, "dup(): %m" );
      exit( 1 );
   }

   if (( stdout = fdopen( conn, "w" )) == NULL )
   {
      syslog( LOG_ERR, "fdopen(): %m" );
      exit( 1 );
   }
}

char *scgi_extract_mime_boundary( char *line )
{
   char **ptr, *ptr1, *ptr2, buffer[ 1024 ], *boundary;
   static char *strings[] = { NULL, NULL, NULL };
   int i;

   if ( line == NULL )
      return NULL;

   for( ptr1 = line; *ptr1; ++ptr1 )
      if ( *ptr1 == '=' )
         break;
      else if ( *ptr1 >= 65 && *ptr1 <= 90 )
         *ptr1 += 32;

   ptr2 = line;

   strings[ 0 ] = "multipart/form-data;";
   strings[ 1 ] = "boundary=";
   strings[ 2 ] = NULL;

   for( ptr = strings; *ptr != NULL; ++ptr )
   {
      ptr1 = *ptr;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( *ptr1 )
         return NULL;

      while( *ptr2 == ' ' || *ptr2 == '\t' || *ptr2 == '\r' || *ptr2 == '\n' )
         ++ptr2;
   }

   if ( *ptr2 == '"' )
   {
      ptr1 = ++ptr2;

      for( i = 0; *ptr2 && *ptr2 != '"'; ++ptr2 )
         ++i;
   }
   else
   {
      ptr1 = ptr2;
      i = 0;

      while( *ptr2 && *ptr2 != ' ' && *ptr2 != '\t' && *ptr2 != '\r' && *ptr2 != '\n' )
      {
         ++ptr2;
         ++i;
      }
   }

   boundary = scgi_str_dup( ptr1, -1 );

   strings[ 0 ] = "--";
   strings[ 1 ] = boundary;
   strings[ 2 ] = NULL;

   for( ; ; )
   {
   AGAIN2:
      if (( line = fgets( buffer, sizeof( buffer ), stdin )) == NULL )
      {
         syslog( LOG_ERR, ( ferror( stdin ) ? "fgets(): %m" : "initial boundary not found in multipart/form-data body" ));
         free( boundary );
         return NULL;
      }

      ptr2 = line;

      for( ptr = strings; *ptr != NULL; ++ptr )
      {
         ptr1 = *ptr;

         while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
         {
            ++ptr1;
            ++ptr2;
         }

         if ( *ptr1 )
            goto AGAIN2;
      }

      ptr1 = strings[ 0 ];

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( ! *ptr1 && ( ! *ptr2 || *ptr2 == '\r' || *ptr2 == '\n' ))
      {
         syslog( LOG_ERR, "empty multipart/form-data document" );
         free( boundary );
         return NULL;
      }

      break;
   }

   return boundary;
}

char **scgi_extract_entity_names( char *line )
{
   char **ptr, *ptr1 = NULL, *ptr2 = NULL, *strings[] = { "content-disposition:", "form-data;", "name=", NULL };
   static char *result[ 2 ];

   for( ptr1 = line; *ptr1; ++ptr1 )
      if ( *ptr1 == ':' )
         break;
      else if ( *ptr1 >= 65 && *ptr1 <= 90 )
         *ptr1 += 32;

   result[ 0 ] = NULL;
   result[ 1 ] = NULL;
   ptr2 = line;

   for( ptr = strings; *ptr != NULL; ++ptr )
   {
      ptr1 = *ptr;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( *ptr1 )
         break;

      while( *ptr2 == ' ' || *ptr2 == '\t' || *ptr2 == '\r' || *ptr2 == '\n' )
         ++ptr2;
   }

   if ( *ptr1 )
      return NULL;

   if ( *ptr2 == '"' )
   {
      ptr1 = ++ptr2;

      while( *ptr2 && *ptr2 != '"' )
         ++ptr2;

      *ptr2 = '\0';
   }
   else
   {
      ptr1 = ptr2;

      while( *ptr2 && *ptr2 != ';' && *ptr2 != ' ' && *ptr2 != '\t' && *ptr2 != '\r' && *ptr2 != '\n' )
         ++ptr2;

      *ptr2 = '\0';
   }

   result[ 0 ] = scgi_str_dup( ptr1, -1 );

   ++ptr2;

   while( *ptr2 == ';' || *ptr2 == ' ' || *ptr2 == '\t' || *ptr2 == '\r' || *ptr2 == '\n' )
      ++ptr2;

   ptr1 = "filename=";

   while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
   {
      ++ptr1;
      ++ptr2;
   }

   if ( ! *ptr1 )
   {
      if ( *ptr2 == '"' )
      {
         ptr1 = ++ptr2;

         while( *ptr2 && *ptr2 != '"' )
            ++ptr2;

         *ptr2 = '\0';
      }
      else
      {
         ptr1 = ptr2;

         while( *ptr2 && *ptr2 != ';' && *ptr2 != ' ' && *ptr2 != '\t' && *ptr2 != '\r' && *ptr2 != '\n' )
            ++ptr2;

         *ptr2 = '\0';
      }

      result[ 1 ] = scgi_str_dup( ptr1, -1 );
   }

   return result;
}

int scgi_is_boundary( char *line, char *boundary )
{
   char *ptr1, *ptr2, *dashes = "--";

   ptr1 = dashes;
   ptr2 = line;

   while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
   {
      ++ptr1;
      ++ptr2;
   }

   if ( ! *ptr1 )
   {
      ptr1 = boundary;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( ! *ptr1 )
      {
         ptr1 = dashes;

         while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
         {
            ++ptr1;
            ++ptr2;
         }

         if ( ! *ptr1 && ( ! *ptr2 || *ptr2 == '\r' || *ptr2 == '\n' ))
            return -1;

         return 1;
      }
   }

   return 0;
}

int scgi_process_mime_data( char *boundary, FILE *file, char **value )
{
   static struct scgi_string *s = NULL, *l = NULL;
   int c, i, result, count = 0, oversize = 0;

   *value = NULL;

   if ( s == NULL )
   {
      s = scgi_make_string();
      l = scgi_make_string();
   }
   else
   {
      s->free += s->used;
      s->used = 0;
      s->top = s->str;

      l->free += l->used;
      l->used = 0;
      l->top = l->str;
   }

   for( ; ; )
   {
      for( ; ; )
      {
         if (( c = fgetc( stdin )) == EOF )
         {
            syslog( LOG_ERR, ( ferror( stdin ) ? "fgetc(): %m" : "premature EOF in multipart/form-data stream" ));
            return 1;
         }

         STRING_APPEND( l, c );

         if ( c == '\n' || l->used == 262144 )
            break;
      }

      if (( result = scgi_is_boundary( l->str, boundary )))
         break;

      if ( ! oversize && s->used && file != NULL )
      {
         if ( ! fwrite( s->str, s->used, 1, file ))
         {
            syslog( LOG_ERR, "fwrite(): %m" );
            return 1;
         }

         s->free += s->used;
         s->top = s->str;
         s->used = 0;
      }

      if ( ! oversize )
      {
         char *ptr;
         
         for( ptr = l->str, i = 0; i < l->used; ++i, ++ptr )
            STRING_APPEND( s, *ptr )

         count += l->used;
      }

      l->free += l->used;
      l->top = l->str;
      l->used = 0;

      if ( ! oversize )
      {
         if ( file != NULL )
         {
            if ( count > scgi_max_file )
               oversize = 1;
         }
         else if ( count > scgi_max_body )
            oversize = 1;
      }
   }

   if ( oversize )
      return 1;

   if ( s->used > 1 )
   {
      s->used -= 2;
      s->top -= 2;
      s->free += 2;
      *s->top = '\0';
   }

   if ( file != NULL )
   {
      if ( s->used && ! fwrite( s->str, s->used, 1, file ))
      {
         syslog( LOG_ERR, "fwrite(): %m" );
         return 1;
      }
   }
   else
      *value = scgi_str_dup( s->str, -1 );

   return ( result == 1 ? 0 : ( result < 0 ? result : 1 ));
}

int scgi_process_mime_entity( char *boundary, char **names, char **value )
{
   char path[ PATH_MAX ];
   int fd, result;
   FILE *file = NULL;

   if ( names[ 1 ] != NULL )
   {
      snprintf( path, sizeof( path ), "%s/XXXXXXXXXXXXX", scgi_root_dir );

      if (( fd = mkstemp( path )) < 0 )
      {
         syslog( LOG_ERR, "mkstemp(): %m" );
         return 1;
      }

      if (( file = fdopen( fd, "w+" )) == NULL )
      {
         syslog( LOG_ERR, "fdopen(): %m" );
         close( fd );
         return 1;
      }
   }

   result = scgi_process_mime_data( boundary, file, value );

   if ( result <= 0 && *value == NULL && file != NULL )
      *value = scgi_str_dup( path, -1 );

   if ( file != NULL )
   {
      fclose( file );

      if ( result > 0 && unlink( path ) < 0 )
         syslog( LOG_ERR, "unlink(): %m" );
   }

   return result;
}

int scgi_process_form_entity( char *boundary, char **name, char **value, char **file )
{
   char buffer[ 1024 ], **names = NULL, *myval;
   int result;

   for( ; ; )
   {
      if ( fgets( buffer, sizeof( buffer ), stdin ) == NULL )
      {
         syslog( LOG_ERR, ( ferror( stdin ) ? "fgets(): %m" : "premature EOF on stdin for multipart/form-data stream" ));

         if ( names != NULL )
         {
            free( names[ 0 ] );

            if ( names[ 1 ] != NULL )
               free( names[ 1 ] );
         }

         return 1;
      }

      if ( *buffer == '\r' || *buffer == '\n' )
         break;

      if ( names == NULL )
         names = scgi_extract_entity_names( buffer );
   }

   if ( names == NULL || ! *names[ 0 ] )
   {
      if ( names != NULL )
      {
         free( names[ 0 ] );

         if ( names[ 1 ] != NULL )
            free( names[ 1 ] );
      }

      syslog( LOG_ERR, "missing or empty name parameter in multipart/form-data entity" );
      return 1;
   }

   /*
    * An empty filename parameter from a file input indicates no file was selected for
    * uploading.
    */

   if ( names[ 1 ] != NULL && ! *names[ 1 ] )
   {
      free( names[ 1 ] );
      names[ 1 ] = NULL;
   }

   result = scgi_process_mime_entity( boundary, names, &myval );

   if ( result <= 0 )
   {
      *name = names[ 0 ];

      if ( names[ 1 ] != NULL )
      {
         *value = names[ 1 ];
         *file = myval;
      }
      else
         *value = myval;
   }
   else
   {
      free( names[ 0 ] );

      if ( names[ 1 ] != NULL )
         free( names[ 1 ] );
   }

   return result;
}

int scgi_process_form_data( char *content_type )
{
   char *boundary, *name = NULL, *value = NULL, *file = NULL, **p_ptr, **f_ptr;
   int result, i, count = 0;

   if (( boundary = scgi_extract_mime_boundary( content_type )) == NULL )
      return 1;

   for( p_ptr = scgi_params, i = 0; i <= MAX_PARAM; ++i )
      *p_ptr++ = NULL;

   for( f_ptr = scgi_files, i = 0; i <= MAX_FILE; ++i )
      *f_ptr++ = NULL;

   p_ptr = scgi_params;
   f_ptr = scgi_files;

   while(( result = scgi_process_form_entity( boundary, &name, &value, &file )) <= 0 )
   {
      if ( ++count > MAX_PARAM )
      {
         if ( name != NULL )
            free( name );
            
         if ( value != NULL )
            free( value );
            
         if ( file != NULL )
         {
            unlink( file );
            free( file );
         }

         if ( result < 0 )
            break;
         else
            continue;
      }
      
      if ( file != NULL )
      {
         *f_ptr++ = name;
         *f_ptr++ = value;
         *f_ptr++ = file;
      }
      else
      {
         *p_ptr++ = name;
         *p_ptr++ = value;
      }
         
      if ( result < 0 )
         break;

      name = value = file = NULL;
   }

   free( boundary );

   return result;
}

char *scgi_find_next( char *ptr )
{
   while( --scgi_header_len )
      if ( ! *ptr++ )
         break;

   if ( ! scgi_header_len )
      return NULL;

   return ptr;
}

int scgi_init_env( int len )
{
   char *ptr;
   int n;

   n = 0;

   for( ptr = scgi_header; ptr != NULL; ptr = scgi_find_next( ptr ))
   {
      scgi_env[ n++ ] = ptr;

      if ( n == MAX_ENV )
         break;
   }

   if ( n % 2 )
      return 1;

   scgi_env[ n ] = NULL;

   return 0;
}

char *scgi_get_env( char *key )
{
   char **ptr, *ptr1, *ptr2;

   for( ptr = scgi_env; *ptr != NULL; ptr += 2 )
   {
      ptr1 = *ptr;
      ptr2 = key;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( ! *ptr1 && ! *ptr2 )
         return *++ptr;
   }

   return NULL;
}

char **scgi_get_envs()
{
   return scgi_env;
}

int scgi_read_header()
{
   char *ptr, buffer[ 64 ];
   int c;

   for( ptr = buffer; ; ++ptr )
   {
      if (( c = getchar()) == EOF )
         return 1;

      if ( c == ':' )
      {
         *ptr = '\0';
         break;
      }

      *ptr = c;
   }

   if (( scgi_header_len = strtol( buffer, NULL, 10 )) <= 0 )
      return 1;

   if (( scgi_header = scgi_memory( scgi_header_len )) == NULL )
      return 1;

   if ( fread( scgi_header, 1, scgi_header_len, stdin ) < c )
   {
      free( scgi_header );
      scgi_header = NULL;
      return 1;
   }

   if ( getchar() != ',' )
   {
      free( scgi_header );
      scgi_header = NULL;
      return 1;
   }

   return scgi_init_env( scgi_header_len );
}

int scgi_read_body()
{
   char *ptr;
   int count;

   if (( ptr = scgi_get_env( "CONTENT_TYPE" )) != NULL &&
         !strncmp( ptr, "multipart/form-data", 19 ))
      return scgi_process_form_data( ptr );

   for( ptr = scgi_header; *ptr; ++ptr )
      ;

   if ( !( count = strtol( ++ptr, NULL, 10 )))
      return 0;

   if ( count < 0 || count > scgi_max_body )
      return 1;

   if (( scgi_body = scgi_memory( count + 1 )) == NULL )
      return 1;

   if ( fread( scgi_body, 1, count, stdin ) < count )
   {
      free( scgi_body );
      scgi_body = NULL;
      return 1;
   }

   scgi_body[ count ] = '\0';

   return 0;
}

void scgi_process_cookies()
{
   char *cookie, *name, *ptr, **ptr2;
   int i;

   for( ptr2 = scgi_cookies, i = 0; i <= MAX_COOKIE; ++i )
      *ptr2++ = NULL;

   cookie = NULL;

   for( ptr2 = scgi_env; ptr2 != NULL; ptr2 += 2 )
      if ( ! strcmp( *ptr2, "HTTP_COOKIE" ))
      {
         cookie = *( ptr2 + 1 );
         break;
      }

   if ( cookie == NULL || ! *cookie )
      return;

   scgi_http_cookie = scgi_str_dup( cookie, -1 );
   cookie = scgi_http_cookie;

   for( i = 0, ptr = cookie; *ptr; ++ptr )
   {
      if ( *ptr == ';' || *ptr == ',' )
      {
         if ( i >= MAX_COOKIE )
            break;

         *ptr = '\0';

         for( name = cookie; *name && *name != '='; ++name )
            ;

         if ( *name )
            *name++ = '\0';

         while( *cookie == ' ' )
            ++cookie;

         scgi_cookies[ i++ ] = cookie;
         scgi_cookies[ i++ ] = name;

         cookie = ptr + 1;
      }
   }

  if ( *cookie && i < MAX_COOKIE - 1 )
  {
      for( name = cookie; *name && *name != '='; ++name )
         ;

      if ( *name )
         *name++ = '\0';

      while( *cookie == ' ' )
         ++cookie;

      scgi_cookies[ i++ ] = cookie;
      scgi_cookies[ i++ ] = name;
   }
}

void scgi_process_params()
{
   char *query, *ptr, *param, **ptr2;
   int i;

   for( ptr2 = scgi_params, i = 0; i <= MAX_PARAM; ++i )
      *ptr2++ = NULL;

   query = NULL;

   for( ptr2 = scgi_env; ptr2 != NULL; ptr2 += 2 )
      if ( ! strcmp( *ptr2, "QUERY_STRING" ))
      {
         query = *( ptr2 + 1 );
         break;
      }

   if ( query == NULL || ! *query )
   {
      if ( scgi_body == NULL )
         return;

      query = scgi_body;
   }
   else
   {
      scgi_query_string = scgi_str_dup( query, -1 );
      query = scgi_query_string;
   }

   for( i = 0, ptr = query; *ptr; ++ptr )
   {
      if ( *ptr == '&' )
      {
         if ( i >= MAX_PARAM )
            break;

         *ptr = '\0';

         for( param = query; *param && *param != '='; ++param )
            ;

         if ( *param )
            *param++ = '\0';

         scgi_params[ i++ ] = scgi_form_decode( query );
         scgi_params[ i++ ] = scgi_form_decode( param );

         query = ptr + 1;
      }
   }

  if ( *query && i < MAX_PARAM - 1 )
  {
      for( param = query; *param && *param != '='; ++param )
         ;

      if ( *param )
         *param++ = '\0';

      scgi_params[ i++ ] = scgi_form_decode( query );
      scgi_params[ i++ ] = scgi_form_decode( param );
   }

   return;
}

char *scgi_get_param( char *param )
{
   char **ptr, *ptr1, *ptr2;

   for( ptr = scgi_params; *ptr != NULL; ptr += 2 )
   {
      ptr1 = *ptr;
      ptr2 = param;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( ! *ptr1 && ! *ptr2 )
         return *( ptr + 1 );
   }

   return NULL;
}

char **scgi_get_params()
{
   return scgi_params;
}

char **scgi_get_file( char *file )
{
   char **ptr, *ptr1, *ptr2;

   for( ptr = scgi_files; *ptr != NULL; ptr += 3 )
   {
      ptr1 = *ptr;
      ptr2 = file;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( ! *ptr1 && ! *ptr2 )
         return ( ptr + 1 );
   }

   return NULL;
}

char **scgi_get_files()
{
   return scgi_files;
}

char *scgi_get_cookie( char *name )
{
   char **ptr, *ptr1, *ptr2;

   for( ptr = scgi_cookies; *ptr != NULL; ptr += 2 )
   {
      ptr1 = *ptr;
      ptr2 = name;

      while( *ptr1 && *ptr2 && *ptr1 == *ptr2 )
      {
         ++ptr1;
         ++ptr2;
      }

      if ( ! *ptr1 && ! *ptr2 )
         return *( ptr + 1 );
   }

   return NULL;
}

char **scgi_get_cookies()
{
   return scgi_cookies;
}

void scgi_set_name( char *name )
{
   char *ptr1, *ptr2;

   if ( name == NULL )
      return;

   scgi_app_name = name;
   ptr2 = scgi_pidfile = scgi_memory( 14 + strlen( name ));

   for ( ptr1 = "/var/run/"; *ptr1; )
      *ptr2++ = *ptr1++;

   for ( ptr1 = name; *ptr1; )
      *ptr2++ = *ptr1++;

   for ( ptr1 = ".pid"; *ptr1; )
      *ptr2++ = *ptr1++;

   *ptr2 = '\0';
}

void scgi_service_connections()
{
   char **ptr;
   int result;

   for( ; ; )
   {
      scgi_busymap[ scgi_slave_no ] = NOT_BUSY;
      scgi_accept_connection();
      scgi_busymap[ scgi_slave_no ] = BUSY;

      if ( scgi_read_header() )
         continue;

      if (( result = scgi_read_body()) > 0 )
      {
         free( scgi_header );
         scgi_header = NULL;
         continue;
      }

      if ( ! result  )
         scgi_process_params();

      scgi_process_cookies();
      scgi_request_handler();

      for( ptr = scgi_params; *ptr != NULL; ++ptr )
         free( *ptr );

      if ( result < 0 )
         for( ptr = scgi_files; *ptr != NULL; ptr += 3 )
         {
            free( *ptr );
            free( *( ptr + 1 ));
            unlink( *( ptr + 2 ));
            free( *( ptr + 2 ));
         }
         
      free( scgi_header );

      if ( scgi_query_string != NULL )
         free( scgi_query_string );

      if ( scgi_http_cookie != NULL )
         free( scgi_http_cookie );

      if ( scgi_body != NULL )
         free( scgi_body );

      scgi_header = scgi_body = scgi_query_string = scgi_http_cookie = NULL;
   }
}

void scgi_init_busymap()
{
   int i;

   if (( scgi_busymap = mmap( 0, scgi_max_slaves, PROT_WRITE | PROT_READ,
                              MAP_ANON | MAP_SHARED, -1, 0 )) == NULL )
   {
      syslog( LOG_ERR, "mmap(): %m" );
      exit( 1 );
   }

   scgi_slaves = scgi_memory( sizeof( pid_t ) * scgi_max_slaves );

   for( i = 0; i < scgi_max_slaves; ++i )
   {
      scgi_busymap[ i ] = EMPTY;
      scgi_slaves[ i ] = 0;
   }
}

void scgi_fork_slave( int i )
{
   pid_t pid;

   switch(( pid = fork()))
   {
      case -1:
         syslog( LOG_ERR, "fork(): %m" );
         exit( 1 );

      case 0:
         scgi_slave_no = i;
         scgi_child_init_func();
         scgi_service_connections();
         exit( 0 );

      default:
         scgi_slaves[ i ] = pid;
   }
}

void scgi_periodic( void ( *func )(), int interval )
{
   scgi_periodic_func = func;

   if ( interval < 0 )
      interval = 0;

   scgi_interval = interval;
   scgi_counter = 0;
}

void scgi_manage_slaves()
{
   pid_t pid;
   int i, idle, last_idle;

   for( ; ; )
   {
      sleep( 1 );

      if ( scgi_interval && scgi_periodic_func != NULL && ++scgi_counter == scgi_interval )
      {
         scgi_periodic_func();
         scgi_counter = 0;
      }

      idle = 0;
      last_idle = -1;

      for( i = 0; i < scgi_max_slaves; ++i )
         if ( scgi_busymap[ i ] == NOT_BUSY )
         {
            last_idle = i;
            ++idle;
         }

      if ( idle > scgi_min_slaves )
      {
         kill( scgi_slaves[ last_idle ], SIGTERM );
         --idle;
      }

      while(( pid = waitpid( -1, NULL, WNOHANG )) > 0 )
      {
         for( i = 0; i < scgi_max_slaves; ++i )
            if ( pid == scgi_slaves[ i ] )
            {
               scgi_slaves[ i ] = 0;
               scgi_busymap[ i ] = EMPTY;
               break;
            }
      }

      while( idle < scgi_min_slaves )
      {
         for( i = 0; i < scgi_max_slaves; ++i )
            if ( ! scgi_slaves[ i ] )
            {
               scgi_fork_slave( i );
               ++idle;
               break;
            }

         if ( i == scgi_max_slaves )
            break;
      }
   }
}

void scgi_set_signals()
{
   struct sigaction sigact;
   int *iptr;
   int sigs[] = {
      SIGPIPE, SIGHUP, SIGQUIT, SIGUSR1, SIGUSR2, SIGALRM, SIGINT, SIGTSTP, -1
   };

   /*
    * We ignore everything but SIGTERM when ! testing.  Otherwise, we
    * catch all the signals, so we may kill our children before exiting.
    * SIGTERM is set to interrupt system calls.
    */

   sigact.sa_handler = scgi_signal_handler;
   sigemptyset( &sigact.sa_mask );
   sigact.sa_flags = 0;

   if ( sigaction( SIGTERM, &sigact, NULL ) < 0 )
      fprintf( stderr, "dombey: sigaction: %s\n", strerror( errno ));

   for( iptr = sigs; *iptr > 0; ++iptr )
      signal( *iptr, ( scgi_testing ? scgi_signal_handler : SIG_IGN ));
}

void scgi_fork_slaves()
{
   int i;

   for( i = 0; i < scgi_min_slaves; ++i )
      scgi_fork_slave( i );
}

void scgi_init_config_file( int argc, char **argv )
{
   int i;

   while(( i = getopt( argc, argv, "xe:l:f:r:m:b:n:p:i:u:g:")) != -1 )
      switch( i )
      {
         case '?':
            continue;

         case 'f':
            scgi_config_file = optarg;
            break;
      }

   optreset = 1;
   optind = 1;
}

int main( int argc, char **argv )
{
   scgi_init_config_file( argc, argv );
   scgi_init_func();

   scgi_set_options( argc, argv );
   scgi_set_signals();

   openlog( scgi_app_name, LOG_PID, LOG_DAEMON );
   scgi_logging = 1;

   if ( !scgi_testing )
      scgi_become_daemon();

   scgi_start_listening();
   scgi_change_identity();
   scgi_init_busymap();

   if ( scgi_testing )
   {
      scgi_child_init_func();
      scgi_service_connections();
   }
   else
   {
      scgi_fork_slaves();
      scgi_manage_slaves();
   }

   return 0;
}
